@@ -23,19 +23,21 @@ from .exceptions import (
     RestMethodNotAllowed,
     NoFiltersInSchema,
     FieldNotInSchema,
-    BadRelatedType,
+    InvalidFieldValue,
     NoResourcesExist,
     MultipleResourcesReturned,
     FilterNotAllowedForField,
     InvalidFieldName,
+    NoDefaultValueInSchema,
     ResourceDeleted,
     CreatedResourceNotFound,
-    BadJsonResponse,
+    ResourceHasNoUri,
+    BadUri,
     ErrorResponse,
     NonExistantResource,
     CannotConnectToAddress,
-    IncorrectEndpointArgs,
-    IncorrectEndpointKwargs,
+    IncorrectNestedResourceArgs,
+    IncorrectNestedResourceKwargs,
 )
 
 from .factory import ResourceFactory
