@@ -16,23 +16,20 @@ import json
 import requests
 
 from .exceptions import (
-    BadJsonResponse,
     ErrorResponse,
     NonExistantResource,
     CannotConnectToAddress,
     ResourceDeleted,
+    IncorrectNestedResourceKwargs,
+    BadUri,
     RestMethodNotAllowed,
-    IncorrectEndpointArgs,
-    IncorrectEndpointKwargs,
 )
 from .schema import TastySchema
-from . import tastytypes
 
 
 class TastyApi(object):
     """Wrap the TastyPie API providing basic get/add/update/delete methods.
 
-    :var max_results: initial value: 100
     :param address: URL of the TastyPie API.
     :type address: str
     """
@@ -45,33 +42,14 @@ class TastyApi(object):
         self._res = None
         self._sess = None
         self.auth = None
-        self.max_results = 100
 
     def _session(self):
         if self._sess is None:
             self._sess = requests.session()
         return self._sess
 
-    def _address(self):
-        return self._addr
-
-    def _resources(self):
-        if self._res is None:
-            self._res = self._transmit(self._session().get, self._address())
-        return self._res
-
-    def _base_url(self):
-        if self._baseurl is None:
-            any_endpoint = next(iter(self._resources().values()))['list_endpoint']
-            excess = any_endpoint.rsplit('/', 2)[0] + '/'
-            self._baseurl = self._address().replace(excess, '')
-        return self._baseurl
-
     def _transmit(self, tx_func, url, params=None, data=None):
-        if params:
-            params = {k: self._encode_for_transmit(v) for k, v in params.items()}
         if data:
-            data = {k: self._encode_for_transmit(v) for k, v in data.items()}
             data = json.dumps(data)
         try:
             response = tx_func(
@@ -86,45 +64,23 @@ class TastyApi(object):
         except (ValueError, TypeError) as err:
             try:
                 if response.text:
-                    args = (response.text, url, params, data)
+                    args = (response.text, err, url, params, data)
                     if 'NotFound: Invalid resource' in response.text:
                         raise NonExistantResource(*args)
-                    if 'MultiValueDictKeyError: ' in response.text:
-                        raise IncorrectEndpointKwargs(*args)
-                    raise BadJsonResponse(*args)
+                    if 'KeyError: ' in response.text:
+                        raise IncorrectNestedResourceKwargs(*args)
+                    raise ErrorResponse(*args)
             # If it was raised before the request was sent, it wasn't a JSON error.
             except UnboundLocalError:
                 raise err
         except requests.exceptions.HTTPError as err:
             if response.status_code == 404 or response.status_code == 410:
                 raise ResourceDeleted(url)
-            raise ErrorResponse(err, response.text, url, params, data)
+            if response.status_code == 405:
+                raise RestMethodNotAllowed(err, url, tx_func)
+            raise ErrorResponse(response.text, err, url, params, data)
         except requests.exceptions.ConnectionError as err:
-            raise CannotConnectToAddress(self._address())
-
-    def _get_resource(self, resource_type):
-        try:
-            return self._resources()[resource_type]['list_endpoint']
-        except KeyError:
-            raise NonExistantResource(resource_type)
-
-    @staticmethod
-    def _encode_for_transmit(obj):
-        # TODO This is getting out of control...
-        if isinstance(obj, list):
-            try:
-                obj = [o.uri() for o in obj]
-            except AttributeError:
-                pass
-        try:
-            obj = obj.strftime(tastytypes.DATETIME_FORMAT1)
-        except AttributeError:
-            pass
-        try:
-            obj = obj.uri()
-        except AttributeError:
-            pass
-        return obj
+            raise CannotConnectToAddress(self.address())
 
     @staticmethod
     def _headers():
@@ -133,69 +89,61 @@ class TastyApi(object):
             'content-type': 'application/json',
         }
 
-    def _resource_url(self, resource):
-        try:
-            resource = resource.uri()
-        except AttributeError:
-            pass
-        return self._base_url() + resource
+    def address(self):
+        """Return the address of the API."""
+        return self._addr
 
-    def _get_endpoint(self, base_url, method_name, *args, **kwargs):
-        args_string = '/'.join(str(a) for a in args)
-        url = '{0}{1}/{2}'.format(base_url, method_name, args_string)
-        if not url.endswith('/'):
-            url += '/'
-        try:
-            return self._transmit(self._session().get, url, params=kwargs)
-        except NonExistantResource as err:
-            raise IncorrectEndpointArgs(*err.args)
+    def create_full_uri(self, uri):
+        """Return the full address of the given URI."""
+        uri_cutoff = len(uri)
+        while uri_cutoff > 2:
+            uri_cutoff -= 1
+            if self.address().endswith(uri[:uri_cutoff]):
+                return self.address() + uri[uri_cutoff:]
+        raise BadUri('Could not find full uri. Address = "{0}", URI = "{1}"'.format(self.address(), uri))
 
-    def get(self, resource_type, schema, **kwargs):
+    def paginate(self, url, **kwargs):
         """Retrieve the objects for a given resource type.
 
         The search can be filtered by passing field=value as kwargs.
 
-        :param resource_type: The TastyPie resource name.
-        :type resource_type: str
+        :param url: The URL of the TastyPie resource.
+        :type url: str
         :param schema: The schema to use for validation.
         :type schema: TastySchema
         :returns: A generator object that yields dicts.
         :rtype dict
         """
-        url = self._base_url() + self._get_resource(resource_type)
-        schema.check_list_request_allowed('get')
         retrieve_all_results = False
         if 'limit' not in kwargs:
-            kwargs.update({'limit': self.max_results})
             retrieve_all_results = True
         result = self._transmit(self._session().get, url, params=kwargs)
         yield result
         # Only continue to retrieve results if the user hasn't specified the
         # number of results to retrieve.
         while result['meta']['next'] and retrieve_all_results:
+            url = self.create_full_uri(result['meta']['next'])
             result = self._transmit(self._session().get, url)
             yield result
 
-    def details(self, resource, schema):
+    def get(self, url, **kwargs):
         """Retrieve the fields for a given URI.
 
-        :param resource: A URI pointing to the TastyPie resource.
+        :param url: The URL of the TastyPie resource.
         :type resource: str
         :param schema: The schema to use for validation.
         :type schema: TastySchema
         :returns: The resource's fields.
         :rtype: dict
         """
-        url = self._resource_url(resource)
-        schema.check_detail_request_allowed('get')
-        return self._transmit(self._session().get, url)
+        return self._transmit(self._session().get, url, params=kwargs)
 
-    def add(self, resource_type, schema, **kwargs):
+    def post(self, url, **kwargs):
         """Add a new resource with the given fields.
 
         The fields can be set by passing in field=value as kwargs.
 
-        :param resource_type: The TastyPie resource name.
+        :param url: The URL of the TastyPie resource.
         :type resource_type: str
         :param schema: The schema to use for validation.
         :type schema: TastySchema
@@ -203,16 +151,14 @@ class TastyApi(object):
             the Resource was defined in the tastypie API (always_return_data).
         :rtype: dict
         """
-        url = self._base_url() + self._get_resource(resource_type)
-        schema.check_list_request_allowed('post')
         return self._transmit(self._session().post, url, data=kwargs) or {}
 
-    def update(self, resource, schema, **kwargs):
-        """Update a given resource with the given fields.
+    def put(self, url, **kwargs):
+        """Put a given resource with the given fields.
 
         The fields can be set by passing in field=value as kwargs.
 
-        :param resource: A URI pointing to the TastyPie resource.
+        :param url: The URL of the TastyPie resource.
         :type resource: str
         :param schema: The schema to use for validation.
         :type schema: TastySchema
@@ -220,55 +166,37 @@ class TastyApi(object):
             the Resource was defined in the tastypie API (always_return_data).
         :rtype: dict
         """
-        url = self._resource_url(resource)
-        try:
-            schema.check_detail_request_allowed('patch')
-            method = self._session().patch
-        except RestMethodNotAllowed:
-            schema.check_detail_request_allowed('put')
-            method = self._session().put
-        return self._transmit(method, url, data=kwargs) or {}
+        return self._transmit(self._session().put, url, data=kwargs) or {}
 
-    def delete(self, resource, schema):
+    def patch(self, url, **kwargs):
+        """Patch a given resource with the given fields.
+
+        The fields can be set by passing in field=value as kwargs.
+
+        :param url: The URL of the TastyPie resource.
+        :type resource: str
+        :param schema: The schema to use for validation.
+        :type schema: TastySchema
+        :returns: Either the resource fields or an empty dict, depending on how
+            the Resource was defined in the tastypie API (always_return_data).
+        :rtype: dict
+        """
+        return self._transmit(self._session().patch, url, data=kwargs) or {}
+
+    def delete(self, url):
         """Remove a given resource from the API.
 
-        :param resource: A URI pointing to the TastyPie resource.
+        :param url: The URL of the TastyPie resource.
         :type resource: str
         :param schema: The schema to use for validation.
         :type schema: TastySchema
         """
-        url = self._resource_url(resource)
-        schema.check_detail_request_allowed('delete')
         self._transmit(self._session().delete, url)
 
-    def delete_all(self, resource_type, schema):
-        """Remove the collection of resources from the API.
-
-        :param resource_type: A URI pointing to the Tastypie resource type.
-        :type resource_type: str
-        :param schema: The schema to use for validation.
-        :type schema: TastySchema
-        """
-        url = self._base_url() + self._get_resource(resource_type)
-        schema.check_list_request_allowed('delete')
-        self._transmit(self._session().delete, url)
-
-    def list_endpoint(self, resource_type, method_name, schema, *args, **kwargs):
-        """Send a GET to a custom endpoint for this resource."""
-        url = self._base_url() + self._get_resource(resource_type)
-        schema.check_list_request_allowed('get')
-        return self._get_endpoint(url, method_name, *args, **kwargs)
-
-    def detail_endpoint(self, resource, method_name, schema, *args, **kwargs):
-        """Send a GET to a custom endpoint for this resource instance."""
-        url = self._resource_url(resource)
-        schema.check_detail_request_allowed('get')
-        return self._get_endpoint(url, method_name, *args, **kwargs)
-
-    def bulk(self, resource_type, schema, resources, delete):
+    def bulk(self, url, schema, resources, delete):
         """Create, update, and delete multiple resources.
 
-        :param resource_type: The TastyPie resource name.
+        :param url: The URL of the TastyPie resource.
         :type resource_type: str
         :param schema: The schema to use for validation.
         :type schema: TastySchema
@@ -277,24 +205,23 @@ class TastyApi(object):
         :param delete: URIs of resources to delete.
         :type delete: list
         """
-        url = self._base_url() + self._get_resource(resource_type)
         schema.check_list_request_allowed('patch')
         data = {'objects': resources, 'deleted_objects': delete}
         # No result is returned in a 202 response.
         self._transmit(self._session().patch, url, data=data)
 
-    def schema(self, resource_type):
+    def schema(self, url):
         """Retrieve the schema for a given resource type.
 
-        :param resource_type: The TastyPie resource name.
+        :param url: The URL of the TastyPie resource.
         :type resource: str
         :returns: A wrapper around the resource's schema.
         :rtype: TastySchema
         :raises: NonExistantResource
         """
         try:
-            url = self._base_url() + self._resources()[resource_type]['schema']
+            url += 'schema/'
             schema_dict = self._transmit(self._session().get, url)
-            return TastySchema(schema_dict, resource_type)
-        except KeyError:
-            raise NonExistantResource(resource_type)
+            return TastySchema(schema_dict, url)
+        except (KeyError, ResourceDeleted):
+            raise NonExistantResource(url)
